import numpy
import pandas
import pymarket
import pytest

@pytest.fixture(autouse=True)
def add_namespace(doctest_namespace):
        doctest_namespace['np'] = numpy
        doctest_namespace['pd'] = pandas
        doctest_namespace['pm'] = pymarket



import pulp
import pandas as pd
from pymarket.bids import BidManager


def maximum_aggregated_utility(bids, *args, reservation_prices=None):
    """Maximizes the total welfare

    Parameters
    ----------
    bids : pd.DataFrame
       Collection of bids

    reservation_prices : dict of floats or None, (Default value = None)
        A maping from user ids to reservation prices. If no reservation
        price for a user is given, his bid will be assumed to be his
        true value.

    Returns
    -------
    status : str
        Status of the optimization problem. Desired output is 'Optimal'
    objective: float
        Maximum aggregated utility that can be obtained
    variables: dict
        A set of values achieving the objective. Maps a pair of bids
        to the quantity traded by them.

    Examples
    ---------

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.add_bid(1.5, 1, 2, False)
    2
    >>> s, o, v = maximum_aggregated_utility(bm.get_df())
    >>> s
    'Optimal'
    >>> o
    2.5
    >>> v
    {(0, 2): 1.0, (1, 2): 0.5}

    If in reality the seller had 0 value for his commodity,
    the social welfare will be 1.5 units larger

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.add_bid(1.5, 1, 2, False)
    2
    >>> rp = {2: 0}
    >>> s, o, v = maximum_aggregated_utility(bm.get_df(),
    ...        reservation_prices=rp)
    >>> s
    'Optimal'
    >>> o
    4.0
    >>> v
    {(0, 2): 1.0, (1, 2): 0.5}
    """

    if reservation_prices is None:
        reservation_prices = {}

    model = pulp.LpProblem("Max aggregated utility", pulp.LpMaximize)
    buyers = bids.loc[bids['buying']].index.values
    sellers = bids.loc[~bids['buying']].index.values
    index = [(i, j) for i in buyers for j in sellers]

    for i, x in bids.iterrows():
        if i not in reservation_prices:
            reservation_prices[i] = x.price

    coeffs = {}
    for x in index:
        coeffs[x] = reservation_prices[x[0]] - reservation_prices[x[1]]

    qs = pulp.LpVariable.dicts('q', index, lowBound=0, cat='Continuous')

    model += pulp.lpSum([qs[x[0], x[1]] * coeffs[x] for x in index])

    for b in buyers:
        model += pulp.lpSum(qs[b, j] for j in sellers) <= bids.iloc[b, 0]

    for s in sellers:
        model += pulp.lpSum(qs[i, s] for i in buyers) <= bids.iloc[s, 0]

    model.solve()

    status = pulp.LpStatus[model.status]
    objective = pulp.value(model.objective)

    variables = {}
    for var in qs:
        varval = qs[var].varValue
        variables[var] = varval

    return status, objective, variables


def percentage_welfare(bids, transactions, reservation_prices=None, **kwargs):
    """Percentage of the total welfare that could be achieved
    calculated based on the transaction lists

    Parameters
    ----------
    bids (pandas dataframe) :
        Table with all the submited bids
    transactions (pandas dataframe) :
        Table with all the transactions that ocurred in the market
    reservation_prices (dict, optional) :
        Reservation prices of the different participants. If None, the bids
        will be assumed to be the truthfull values.

    Returns
    -------
    ratio : float
        The ratio of the maximum social welfare achieved by the
        collection of transactions.

    Examples
    ---------
    Only bid 0 and 2 trade. That represents a net utility of 2
    which is 80% of the total max utility 2.5

    >>> tm = pm.TransactionManager()
    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.add_bid(1.5, 1, 2, False)
    2
    >>> tm.add_transaction(0, 1, 2, 2, False)
    0
    >>> tm.add_transaction(2, 1, 2, 0, False)
    1
    >>> percentage_welfare(bm.get_df(), tm.get_df())
    0.8
    """
    reservation_prices = {}
    for i, x in bids.iterrows():
        if i not in reservation_prices:
            reservation_prices[i] = x.price

    _, objective, _ = maximum_aggregated_utility(bids, reservation_prices)

    tmp = bids.reset_index().rename(columns={'index': 'bid'})
    tmp = tmp[['bid', 'price', 'buying']]
    merged = transactions.merge(tmp, on='bid')

    buyers = merged.loc[merged['buying']]
    profit_buyers = (buyers.price_y - buyers.price_x) * buyers.quantity
    profit_buyers = profit_buyers.sum()

    sellers = merged.loc[~merged['buying']]
    profit_sellers = (sellers.price_x - sellers.price_y) * sellers.quantity
    profit_sellers = profit_sellers.sum()

    welfare = profit_buyers + profit_sellers

    if objective > 0:
        return welfare / objective
    else:
        return None



import pandas as pd
import numpy as np


def calculate_profits(
        bids,
        transactions,
        reservation_prices=None,
        fees=None,
        **kwargs):
    """Extras from the transactions and the bids the profit
    of each player and the market maker

    Parameters
    ----------
    bids : pd.DataFrame
        Collections of bids to be used

    transactions : pd.DataFrame
        Collection of transactions to be taken into account
    reservation_prices : dict, (Default value = None)
        Maping between users and their reservation prices. If None,
        it is assumed that each user bided truthfully and the
        information is extracted from the bid.
    fees : np.ndarray, (Default value = None)
        List of fees that each user has to pay to trade in the market.

    Returns
    -------
    profit: dict
        A dictionary with three values:
        * player_bid: A list with the profits of each user
        using their bids as reservation prices
        * player_reservation: Same as above but using
        their reservation prices, if none are provided,
        it is the same as `player_bid`
        * market: profit of the market maker


    Examples
    ---------
    >>> tm = pm.TransactionManager()
    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.add_bid(1.5, 1, 2, False)
    2
    >>> tm.add_transaction(0, 1, 2, 2, False)
    0
    >>> tm.add_transaction(2, 1, 2, 0, False)
    1
    >>> rp = {2: 0}
    >>> profits = calculate_profits(bm.get_df(), tm.get_df(),
    ...        reservation_prices=rp)
    >>> profits['player_bid']
    array([1, 0, 1])
    >>> profits['player_reservation']
    array([1, 0, 2])
    >>> profits['market']
    0.0
    """
    users = sorted(bids.user.unique())
    buyers = bids.loc[bids['buying']].index.values
    sellers = bids.loc[~bids['buying']].index.values

    if reservation_prices is None:
        reservation_prices = {}
    for i, x in bids.iterrows():
        if i not in reservation_prices:
            reservation_prices[i] = x.price

    if fees is None:
        fees = np.zeros(bids.user.unique().shape[0])

    profit = {}
    for case in ['bid', 'reservation']:
        tmp = bids.reset_index().rename(columns={'index': 'bid'}).copy()
        tmp = tmp[['bid', 'price', 'buying', 'user']]
        if case == 'reservation':
            tmp.price = tmp.apply(
                lambda x: reservation_prices.get(x.bid, x.price), axis=1)
        merged = transactions.merge(tmp, on='bid').copy()
        merged['gain'] = merged.apply(lambda x: get_gain(x), axis=1)
        profit_player = merged.groupby('user')['gain'].sum()
        # print(profit_player)
        profit_player = np.array([profit_player.get(x, 0) for x in users])
        profit[f'player_{case}'] = profit_player

        if case == 'bid':
            # print(merged)
            mb = merged.loc[merged['buying']]
            ms = merged.loc[~merged['buying']]
            # print(ms)
            # print(ms.quantity.sum(), mb.quantity.sum())
            # print(ms.price_x * ms.quantity)
            profit_market = (mb.price_x * mb.quantity).values.sum()
            profit_market -= (ms.price_x * ms.quantity).values.sum()
            profit_market += fees.sum()
            profit['market'] = profit_market

    return profit


def get_gain(row):
    """Finds the gain of the row

    Parameters
    ----------
    row : pandas row
       Row obtained by merging a transaction with a
       bid dataframe

    Returns
    -------
    gain
        The gain obtained by the row

    """
    gap = row.price_y - row.price_x
    if not row.buying:
        gap = - gap
    return gap * row.quantity



import matplotlib.pyplot as plt
import numpy as np
from pymarket import Market

"""
Fake test to compensate coverage but do not actually check
the produced image automatically
"""


def test_plots_FAKE():
    """
    Tests the image generated by plot huang against
    a visualy correct file
    """
    r = np.random.RandomState(1234)
    mar = Market()
        
    mar.accept_bid(1, 6.7, 0, True, 0)
    mar.accept_bid(1, 6.6, 1, True, 0)
    mar.accept_bid(1, 6.5, 2, True, 0)
    mar.accept_bid(1, 6.4, 3, True, 0)
    mar.accept_bid(1, 6.3, 4, True, 0)
    mar.accept_bid(1, 6, 5, True, 0)

    mar.accept_bid(1, 1, 6, False, 0)
    mar.accept_bid(1, 2, 7, False, 0)
    mar.accept_bid(2, 3, 8, False, 0)
    mar.accept_bid(2, 4, 9, False, 0)
    mar.accept_bid(1, 6.1, 10, False, 0)

    bids = mar.bm.get_df()
    transactions, extras = mar.run('huang')
    mar.plot_method('huang')

    transactions, extras = mar.run('p2p', r=r)
    mar.plot_method('p2p')

    transactions, extras = mar.run('muda', r=r)
    mar.plot_method('muda')


from pymarket.bids import BidManager
from pymarket.mechanisms import *
from pymarket.transactions import TransactionManager
from pymarket.statistics import *
from pymarket.plot import plot_demand_curves, plot_trades_as_graph, plot_both_side_muda, plot_huang_auction

MECHANISM = {
    'huang': HuangAuction,
    'muda': MudaAuction,
    'p2p': P2PTrading,
}

STATS = {

    'percentage_traded': percentage_traded,
    'percentage_welfare': percentage_welfare,
    'profits': calculate_profits,
}


class Market():

    """General interface for calling the different
    market mechanisms

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self):
        """TODO: to be defined1."""
        self.bm = BidManager()

    def accept_bid(self, *args):
        """Adds a bid to the bid manager

        Parameters
        ----------
        bid : TODO
            
        *args :
            

        Returns
        -------

        
        """
        bid_id = self.bm.add_bid(*args)
        return bid_id

    def run(self, algo, *args, **kwargs):
        """Runs a given mechanism with the current
        bids

        Parameters
        ----------
        algo : TODO
            
        *args :
            
        **kwargs :
            

        Returns
        -------

        
        """
        df = self.bm.get_df()
        mec = MECHANISM[algo](df, *args, **kwargs)
        transactions, extra = mec.run()
        self.transactions = transactions
        self.extra = extra
        return transactions, extra

    def statistics(self, reservation_prices=None, exclude=[]):
        """Computes the standard statistics of the market

        Parameters
        ----------
        reservation_prices (dict, optional) :
            the reservation prices of the users. If there is none,
            the bid will be assumed truthfull
        reservation_prices :
             (Default value = None)
        exclude :
            List of mechanisms to ignore will comuting statistics

        Returns
        -------
        stats : dict
            Dictionary with the differnt statistics. Currently:
                * percentage_welfare
                * percentage_traded
                * profits

        
        """
        stats = {}
        extras = {}
        if 'fees' in self.extra:
            extras['fees'] = self.extra['fees']
        extras['reservation_prices'] = reservation_prices
        for stat in STATS:
            if stat not in exclude:
                stats[stat] = STATS[stat](
                    self.bm.get_df(),
                    self.transactions.get_df(),
                    **extras
                )
        self.stats = stats
        return stats

    def plot(self):
        """Plots both demand curves"""
        df = self.bm.get_df()
        plot_demand_curves(df)

    def plot_method(self, method, ax=None):
        """Plots all trades as a bipartite graph.
        It makes sense only for P2P

        Parameters
        ----------
        method :
            
        ax :
             (Default value = None)

        Returns
        -------

        
        """

        trans = self.transactions
        bids = self.bm
        e = self.extra
        if method == 'p2p':
            ax = plot_trades_as_graph(bids, trans, ax)
        elif method == 'muda':
            
            left_players = e['left']
            right_players = e['right']
            left_price = e['price_left']
            right_price = e['price_right']
            ax = plot_both_side_muda(
                bids,
                left_players,
                right_players,
                left_price,
                right_price)
        elif method == 'huang':
            price_sell = e['price_sell']
            price_buy = e['price_buy']
            quantity_traded = e['quantity_traded']
            ax = plot_huang_auction(
                bids,
                price_sell,
                price_buy,
                quantity_traded,
                ax
            )

        return ax



import pandas as pd
import numpy as np


class TransactionManager:
    """
    An interaface to store and
    manage all transactions.
    Transactions are the minimal unit to represent
    the outcome of a market.

    Attributes
    -----------
    name_col: list of str
        Name of the columns to use in the dataframe
        returned.
    n_trans: int
        Number of transactions currently in the Manager
    trans: list of tuples
        List of the actual transactions available
    """

    name_col = ['bid', 'quantity', 'price', 'source', 'active']

    def __init__(self):
        """
        """
        self.n_trans = 0
        self.trans = []

    def add_transaction(self, bid, quantity, price, source, active):
        """Add a transaction to the transactions list

        Parameters
        ----------
        bid : int
            Unique identifier of the bid
        quantity : float
            transacted quantity
        price : float
            transacted price
        source : int
            Identifier of the second party in the trasaction,
            -1 if there is no clear second party, such as
            in a double auction.
        active :
            True` if the bid is still active after the
            transaction.

        Returns
        --------
        trans_id: int
            id of the added transaction, -1 if fails

        Examples
        ---------

        >>> tm = pm.TransactionManager()
        >>> tm.add_transaction(1, 0.5, 2.1, -1, False)
        0
        >>> tm.trans
        [(1, 0.5, 2.1, -1, False)]
        >>> tm.n_trans
        1
        """

        new_trans = (bid, quantity, price, source, active)
        self.trans.append(new_trans)
        self.n_trans += 1

        return self.n_trans - 1

    def get_df(self):
        """Returns the transaction dataframe

        Parameters
        ----------

        Returns
        -------
        df: pd.DataFrame
            A pandas dataframe representing all the transactions
            stored.

        Examples
        ---------
        >>> tm = pm.TransactionManager()
        >>> tm.add_transaction(1, 0.5, 2.1, -1, False)
        0
        >>> tm.add_transaction(5, 0, 0, 3, True)
        1
        >>> tm.get_df()
           bid  quantity  price  source  active
        0    1       0.5    2.1      -1   False
        1    5       0.0    0.0       3    True
        """

        df = pd.DataFrame(self.trans, columns=self.name_col)
        return df

    def merge(self, other):
        """
        Merges two transaction managers with each other
        There are no checks on whether the new
        TransactionManger is consisten after the
        merge.

        Parameters
        ----------
        other : TransactionManager
            A different transaction manager to merge
            with

        Returns
        -------
        trans : TransactionManager
            A new transaction Manager
            with the transactions of the two.

        Examples
        ---------
        >>> tm_1 = pm.TransactionManager()
        >>> tm_1.add_transaction(1, 0.5, 2.1, -1, False)
        0
        >>> tm_2 = pm.TransactionManager()
        >>> tm_2.add_transaction(5, 0, 0, 3, True)
        0
        >>> tm_3 = tm_1.merge(tm_2)
        >>> tm_3.get_df()
           bid  quantity  price  source  active
        0    1       0.5    2.1      -1   False
        1    5       0.0    0.0       3    True

        """

        assert isinstance(other, TransactionManager)

        trans = TransactionManager()
        for t in self.trans:
            trans.add_transaction(*t)
        for t in other.trans:
            trans.add_transaction(*t)

        return trans



# -*- coding: utf-8 -*-

import pymarket.bids
import pymarket.mechanisms
import pymarket.statistics
import pymarket.transactions
import pymarket.plot
import pymarket.utils
import pymarket.datasets

from pymarket.market import *

"""Topymarket.p-level package for pymarket."""

__author__ = """Diego Kiedanki"""
__email__ = 'gusok@protonmail.com'
__version__ = '0.7.5'



"""
Some processing functions to deal with transactions
"""

import pandas as pd

from pymarket.transactions.transactions import TransactionManager


def split_transactions_merged_players(transactions, bids, maping, fees=None):
    """
    Splits the transactions of a market that used merged bids into the original
    bids
    Uses a proportional split, based on the offered (or asked) quantity by
    each player.

    Parameters
    ----------
    transactions: TransactionManager
        the transactions manager returned by the mechanism.
    bids: pandas dataframe
        the original bid dataframe where some players might be repeated
    maping: pandas dataframe
        A maping between the bids in the transaction dataframe and the original
        bids.

    Returns
    --------
    transactions_splited: pandas dataframe
        the result of splitting each merged bid in the transactions
        dataframe
    fees: dict or None
        dictionary obtained by splitting the fees equal to the transactions

    Examples
    -----------
    >>> bm = pm.BidManager()
    >>> tm = pm.TransactionManager()
    >>> bm.add_bid(1, 1, 0)
    0
    >>> bm.add_bid(2, 1, 1)
    1
    >>> tm.add_transaction(0, 1, 1, -1, False)
    0
    >>> tm_2 = split_transactions_merged_players(tm, bm.get_df(), {0:[0,1]})
    >>> tm_2.get_df()
       bid  quantity  price  source  active
    0    0  0.333333      1      -1   False
    1    1  0.666667      1      -1   False

    """

    trans = TransactionManager()
    new_rows = []
    for i, t in transactions.get_df().iterrows():
        rows = maping[t.bid]
        if True:  # TODO make more options available
            perc = bids.iloc[rows, :] / bids.iloc[rows, :].quantity.sum()
            perc = perc.quantity
        if len(rows) > 1 and fees is not None:
            fee = fees.pop(t.user, None)
        for r in rows:
            #t_ = pd.DataFrame(t).copy().T
            # print(r)
            t_ = t.copy().values
            # print(t_)
            t_[0] = r
            t_[1] *= perc[r]
            if len(rows) > 1 and fees is not None and fee is not None:
                fees[bids.iloc[r, :].user] = fee * perc[r]
            #t_.iloc[0, 0] = r
            #t_.iloc[0, 1] *= perc[r]
            # new_rows.append(t_)
            trans.add_transaction(*t_)
    #transactions_splited = pd.concat(new_rows)
    if fees is not None:
        return trans, fees
    else:
        return trans



import networkx as nx
import pandas as pd
import matplotlib.pyplot as plt


def plot_trades_as_graph(
        bids: pd.DataFrame,
        transactions: pd.DataFrame,
        ax=None):
    """Plots all the bids as a bipartit graph
    with buyers and trades and an edge between
    each pair that traded

    Parameters
    ----------
    bids
        Collection of bids to be used
    transactions
        Collection of transactions to be used
    ax
        The axe in which the figure should be ploted

    Returns
    -------
     axe : matplotlib.axes._subplots.AxesSubplot
        The axe in which the figure was plotted.

    """
    bids = bids.get_df()
    tmp = transactions.get_df()
    tmp['user_1'] = tmp.bid.map(bids.user)
    tmp['user_2'] = tmp.source.map(bids.user)
    tmp['buying'] = tmp.bid.map(bids.buying)
    buyers = bids.loc[bids['buying']].index.values

    G = nx.from_pandas_edgelist(tmp, 'user_1', 'user_2')

    edge_labels = {}
    duplicated_labels = tmp.set_index(
        ['user_1', 'user_2'])['quantity'].to_dict()
    for (x, y), v in duplicated_labels.items():
        if ((x, y) not in edge_labels and (y, x) not in edge_labels):
            edge_labels[(x, y)] = v

    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 6))

    pos = nx.bipartite_layout(G, buyers, align='horizontal', scale=3)
    _ = nx.draw_networkx_nodes(
        G,
        pos=pos,
        ax=ax,
        node_color='k',
        node_size=500)
    _ = nx.draw_networkx_labels(G, pos=pos, ax=ax, font_color='w')
    _ = nx.draw_networkx_edges(G, pos=pos, label=G, ax=ax)
    _ = nx.draw_networkx_edge_labels(
        G,
        pos=pos,
        edge_labels=edge_labels,
        label_pos=0.9,
        ax=ax)
    _ = ax.axis('off')

    return ax



import numpy as np
from pymarket import Market


def test_market_init():
    """Tests that the market can be initialized
    and bids added
    Returns
    -------
    TODO

    """

    # Initialize the dataset
    mar = Market()

    mar.accept_bid(1, 4, 0, True, 0)
    mar.accept_bid(1, 3, 1, True, 0)
    mar.accept_bid(1, 2, 2, True, 0)
    mar.accept_bid(1, 1, 3, True, 0)
    mar.accept_bid(1, 1, 10, False, 0)
    mar.accept_bid(1, 2, 11, False, 0)
    mar.accept_bid(1, 3, 12, False, 0)
    
    algos = ['huang', 'muda', 'p2p']
    
    trans = {}
    extras = {}
    stats = {}
    ## Run the three algorithms
    for al in algos:
        if al != 'huang':
            r = np.random.RandomState(1234)
            tr, ex = mar.run(al, r=r)
            st = mar.statistics()
        else:
            tr, ex = mar.run(al)
            st = mar.statistics()
        trans[al] = tr
        extras[al] = ex
        stats[al] = st
    
    for al in algos:
        print(al, extras[al])
        print(al, trans[al].get_df())
        print(al, stats[al])
    #### Test MUDA
    ####################

    true_trans = np.array([
       [4 , 1.0 , 3.0 , -1 , False] ,
       [1 , 1.0 , 3.0 , -1 , False] ,
       [5 , 1.0 , 2.0 , -1 , False] ,
       [0 , 1.0 , 2.0 , -1 , False] ,
        ]).astype(float)

    res_trans = trans['muda'].get_df().astype(float)
    assert np.allclose(true_trans, res_trans)
    assert np.allclose(stats['muda']['percentage_traded'], 2/3)
    assert np.allclose(stats['muda']['percentage_welfare'], 1)
    ###################3


    #### Test HUANG
    ###################

    
    true_trans = np.array([
       [4 , 1.0 , 2.0 , -1 , False] ,
       [0 , 1.0 , 3.0 , -1 , False] ,
        ]).astype(float)

    res_trans = trans['huang'].get_df().astype(float)
    assert np.allclose(true_trans, res_trans)
    assert np.allclose(stats['huang']['percentage_traded'], 1/3)
    assert np.allclose(stats['huang']['percentage_welfare'], 1/2)
    #### Test P2P
    ###################



    true_trans = np.array([
       [1 , 1.0 , 2.0 , 4 , False] ,
       [4 , 1.0 , 2.0 , 1 , False] ,
       [3 , 0   , 0   , 6 , True]  ,
       [6 , 0   , 0   , 3 , True]  ,
       [0 , 1   , 3   , 5 , False] ,
       [5 , 1   , 3   , 0 , False] ,
       [2 , 0   , 0   , 6 , True]  ,
       [6 , 0   , 0   , 2 , True]  ,

        ]).astype(float)

    res_trans = trans['p2p'].get_df().astype(float)
    assert np.allclose(true_trans, res_trans)
    assert np.allclose(stats['p2p']['percentage_traded'], 2/3)
    assert np.allclose(stats['p2p']['percentage_welfare'], 1)


    ##################



import pytest
import numpy as np
from pymarket.bids import *

def test_demand_curve_from_bids_0(bid_dataset_0):
    df = bid_dataset_0
    demand_curve, order = demand_curve_from_bids(df)
    
    data = np.array([
       [2, 4],
       [3, 3],
       [8, 1],
       [np.inf, 0]
        ])


    assert np.allclose(demand_curve, data)
    assert np.allclose(order, np.array([1, 0, 2]))

    xs = [-1   , 0 , 1 , 2 , 2.5 , 3 , 4 , 8 , 1000 , np.inf]
    ys = [None , 4 , 4 , 4 , 3   , 3 , 1 , 1 , 0    , 0]

    for x, y in zip(xs, ys):
        f_ = get_value_stepwise(x, demand_curve)
        if y is not None:
            assert np.allclose(f_, y)
        else:
            assert y == f_

def test_supply_curve_from_bids_0(bid_dataset_0):
    df = bid_dataset_0
    demand_curve, order = supply_curve_from_bids(df)
    
    data = np.array([
       [1, 1],
       [5, 2],
       [10, 6],
       [np.inf, np.inf]
        ])


    assert np.allclose(demand_curve, data)
    assert np.allclose(order, np.array([4, 3, 5]))

    xs = [-1   , 0 , 1 , 2 , 2.5 , 5 , 7 , 10 , 10.001 , np.inf]
    ys = [None , 1 , 1 , 2 , 2   , 2 , 6 , 6  , np.inf , np.inf]

    for x, y in zip(xs, ys):
        f_ = get_value_stepwise(x, demand_curve)
        if y is not None:
            assert np.allclose(f_, y)
        else:
            assert y == f_


def test_equilibrium_quantity_0(bid_dataset_0):
    df = bid_dataset_0

    demand_curve, _ = demand_curve_from_bids(df)
    supply_curve, _ = supply_curve_from_bids(df)
    
    equilibrium_quantity, demand_index, supply_index, price = intersect_stepwise(
            demand_curve, supply_curve)

    assert np.allclose(equilibrium_quantity, 3)
    assert np.allclose(demand_index, 1)
    assert np.allclose(supply_index, 1)
    assert np.allclose(price, 2)

def test_equilibrium_quantity_1(bid_dataset_1):
    df = bid_dataset_1

    demand_curve, _ = demand_curve_from_bids(df)
    supply_curve, _ = supply_curve_from_bids(df)
    
    equilibrium_quantity, demand_index, supply_index, price = intersect_stepwise(
            demand_curve, supply_curve)

    assert np.allclose(equilibrium_quantity, 6)
    assert np.allclose(demand_index, 5)
    assert np.allclose(supply_index, 3)
    assert np.allclose(price, 5)



from functools import wraps
import numpy as np

PRECISION = 8


def check_equal_price(f):
    """CHeck wheather there are two bids
    with the same price in the same side
    and in that case rises an error

    Parameters
    ----------
    f  :(function, mechanisms)
        Mechanisms to be tested


    Returns
    -------

    """
    @wraps(f)
    def wrapper(*args, **kwds):
        """

        Parameters
        ----------
        *args :

        **kwds :


        Returns
        -------

        """
        bids = args[0]

        buy = bids.loc[bids['buying'], :]
        sell = bids.loc[~bids['buying'], :]
        dfs = [buy, sell]

        for df in dfs:
            count = df.groupby(['price', 'user'])['user'].count()
            count = count > 1
            if count.any():
                raise NotImplementedError
        return f(*args, **kwds)
    return wrapper



import pandas as pd
import networkx as nx
import numpy as np
from typing import Union
from pymarket.transactions import TransactionManager
from pymarket.bids import BidManager
from pymarket.mechanisms import Mechanism, MechanismReturn

RandomState = Union[np.random.RandomState, None]


def p2p_random(bids: pd.DataFrame, p_coef: float=0.5,
               r: RandomState=None) -> MechanismReturn:
    """Computes all the trades using a P2P random trading
    process inspired in [1].

    Parameters
    ----------
    bids
        Collection of bids that will trade.
        Precondition: a user participates only in one
        side of the market, i.e, it cannot sell and buy in
        the same run.
    p_coef
        coefficient to calculate the trading price as a convex
        combination of the price of the seller and the price of
        the buyer. If 1, the seller gets all the profit and if 0,
        the buyer gets all the profit.
    r
        Random state to generate stochastic values. If None,
        then the outcome of the market will be different on
        each run.

    Returns
    -------
    trans : TransactionManger
        Collection of all the transactions that ocurred in the market

    extra : dict
        Extra information provided by the mechanisms.
        Keys:

        * trading_list: list of list of tuples of all the pairs that traded in each round.

    Notes
    -------
    [1] Blouin, Max R., and Roberto Serrano. "A decentralized market with
    common values uncertainty: Non-steady states." The Review of Economic
    Studies 68.2 (2001): 323-346.

    Examples
    ---------

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 0.5, 1)
    1
    >>> bm.add_bid(1, 1, 2, False)
    2
    >>> bm.add_bid(1, 2, 3, False)
    3
    >>> r = np.random.RandomState(420)
    >>> trans, extra = p2p_random(bm.get_df(), r=r)
    >>> extra
    {'trading_list': [[(0, 3), (1, 2)]]}
    >>> trans.get_df()
       bid  quantity  price  source  active
    0    0         1    2.5       3   False
    1    3         1    2.5       0   False
    2    1         0    0.0       2    True
    3    2         0    0.0       1    True

    """
    r = np.random.RandomState() if r is None else r
    trans = TransactionManager()
    buying = bids[bids.buying]
    selling = bids[bids.buying == False]
    Nb, Ns = buying.shape[0], selling.shape[0]

    quantities = bids.quantity.values.copy()
    prices = bids.price.values.copy()

    inactive_buying = []
    inactive_selling = []

    # Enumerate all possible trades
    pairs = np.ones((Nb + Ns, Nb * Ns), dtype=bool)
    pairs_inv = []
    i = 0
    for b in buying.index:
        for s in selling.index:
            pairs[b, i] = False  # Row b has 0s whenever the pair involves b
            pairs[s, i] = False  # Same for s
            pairs_inv.append((b, s))
            i += 1

    active = np.ones(Nb * Ns, dtype=bool)
    tmp_active = active.copy()
    general_trading_list = []
    # Loop while there is quantities to trade or not all
    # possibilities have been tried
    while quantities.sum() > 0 and tmp_active.sum() > 0:
        trading_list = []
        while tmp_active.sum() > 0:  # We can select a pair
            where = np.where(tmp_active == 1)[0]
            x = r.choice(where)
            trade = pairs_inv[x]
            active[x] = False  # Pair has already traded
            trading_list.append(trade)
            tmp_active &= pairs[trade[0], :]  # buyer and seller already used
            tmp_active &= pairs[trade[1], :]

        general_trading_list.append(trading_list)
        for (b, s) in trading_list:
            if prices[b] >= prices[s]:
                q = min(quantities[b], quantities[s])
                p = prices[b] * p_coef + (1 - p_coef) * prices[s]
                trans_b = (b, q, p, s, (quantities[b] - q) > 0)
                trans_s = (s, q, p, b, (quantities[s] - q) > 0)
                quantities[b] -= q
                quantities[s] -= q
            else:
                trans_b = (b, 0, 0, s, True)
                trans_s = (s, 0, 0, b, True)
            trans.add_transaction(*trans_b)
            trans.add_transaction(*trans_s)

        inactive_buying = [b for b in buying.index if quantities[b] == 0]
        inactive_selling = [s for s in selling.index if quantities[s] == 0]

        tmp_active = active.copy()
        for inactive in inactive_buying + inactive_selling:
            tmp_active &= pairs[inactive, :]

    extra = {'trading_list': general_trading_list}
    return trans, extra


class P2PTrading(Mechanism):

    """Interface for P2PTrading.

    Parameters
    -----------

    bids: pd.DataFrame
        Collections of bids to use

    """

    def __init__(self, bids, *args, **kwargs):
        """
        """
        Mechanism.__init__(self, p2p_random, bids, *args, **kwargs)



"""
Implements processing techniques applied to bids before
mechanisms can use them
"""
import numpy as np
import pandas as pd
from typing import Callable, List
from pymarket.bids import BidManager


def new_player_id(index: int) -> Callable[[List[int]], int]:
    """Helper function for merge_same_price.
    Creates a function that returns consecutive integers.

    Parameters
    -----------
    index
        First identifier to use for the
        new fake players

    Returns
    -------
    Callable
        Function that maps a list
        of user ids into a new user id.

    Examples
    ----------
    >>> id_gen = new_player_id(6)
    >>> id_gen([3])
    3
    >>> id_gen([5])
    5
    >>> id_gen([0, 1])
    6
    >>> id_gen([2, 4])
    7

    """

    def new_id(users: List[int]) -> int:
        """
        Generates a unique identifier for a
        list of users. If the list has
        only one user, then the id is mantained
        else, a new user id is created for the
        whole list.

        Parameters
        ----------
        users
            List of 1 or more user's identifiers.
            Precondition: all elements of users
            are smaller than index.
        Returns
        -------
        int
            The old identifier if in the list
            there was only one player
            and or the next new consecutive
            identifier if there where more
            than one.

        """
        nonlocal index
        if len(users) > 1:
            new_index = index
            index += 1
        else:
            new_index = users[0]

        return new_index

    return new_id


def merge_same_price(df: pd.DataFrame, prec: float=5) -> pd.DataFrame:
    """
    Process a collection of bids by merging in each
    side (buying or selling) all players with the same
    price into a new user with their aggregated quantity

    Parameters
    ----------
    df
        Collection of bids to process
    prec
        Number of digits to use after the comma
        while comparing floating point prices
        as equal.

    Returns
    -------
    dataframe_new: pd.DataFrame
        The new collection of bids where
        players with the same price have
        been merged into one.

    final_maping : dict
        Maping from new bids index to the
        old bids index.

    Examples
    ---------
    >>> bm = BidManager()
    >>> bm.add_bid(0.3, 1, 0)
    0
    >>> bm.add_bid(0.7, 1, 1)
    1
    >>> bm.add_bid(2, 1, 2, False)
    2
    >>> bm.add_bid(1, 2.444446, 3, False)
    3
    >>> bm.add_bid(3, 2.444447, 4, False)
    4
    >>> bm.get_df()
       quantity     price  user  buying  time  divisible
    0       0.3  1.000000     0    True     0       True
    1       0.7  1.000000     1    True     0       True
    2       2.0  1.000000     2   False     0       True
    3       1.0  2.444446     3   False     0       True
    4       3.0  2.444447     4   False     0       True
    >>> bids, index = pm.merge_same_price(bm.get_df(), 5)
    >>> bids
       quantity    price  user  buying  time  divisible
    0       1.0  1.00000     5    True     0       True
    1       2.0  1.00000     2   False     0       True
    2       4.0  2.44445     6   False     0       True
    >>> index
    {0: [0, 1], 1: [2], 2: [3, 4]}


    >>> mar = pm.Market()
    >>> mar.accept_bid(250, 200, 0, True) # CleanRetail
    0
    >>> mar.accept_bid(300, 110, 1, True) # El4You
    1
    >>> mar.accept_bid(120, 100, 2, True) # EVcharge
    2
    >>> mar.accept_bid( 80,  90, 3, True) # QualiWatt
    3
    >>> mar.accept_bid( 40,  85, 4, True) # IntelliWatt
    4
    >>> mar.accept_bid( 70,  75, 1, True) # El4You
    5
    >>> mar.accept_bid( 60,  65, 0, True) # CleanRetail
    6
    >>> mar.accept_bid( 45,  40, 4, True) # IntelliWatt
    7
    >>> mar.accept_bid( 30,  38, 3, True) # QualiWatt
    8
    >>> mar.accept_bid( 35,  31, 4, True) # IntelliWatt
    9
    >>> mar.accept_bid( 25,  24, 0, True) # CleanRetail
    10
    >>> mar.accept_bid( 10,  21, 1, True) # El4You
    11

    >>> mar.accept_bid(120,   0, 5, False) # RT
    12
    >>> mar.accept_bid(50,    0, 6, False) # WeTrustInWind
    13
    >>> mar.accept_bid(200,  15, 7, False) # BlueHydro
    14
    >>> mar.accept_bid(400,  30, 5, False) # RT
    15
    >>> mar.accept_bid(60, 32.5, 8, False) # KøbenhavnCHP
    16
    >>> mar.accept_bid(50,   34, 8, False) # KøbenhavnCHP
    17
    >>> mar.accept_bid(60,   36, 8, False) # KøbenhavnCHP
    18
    >>> mar.accept_bid(100,37.5, 9, False) # DirtyPower
    19
    >>> mar.accept_bid(70,   39, 9, False) # DirtyPower
    20
    >>> mar.accept_bid(50,   40, 9, False) # DirtyPower
    21
    >>> mar.accept_bid(70,   60, 5, False) # RT
    22
    >>> mar.accept_bid(45,   70, 5, False) # RT
    23
    >>> mar.accept_bid(50,  100, 10, False) # SafePeak
    24
    >>> mar.accept_bid(60,  150, 10, False) # SafePeak
    25
    >>> mar.accept_bid(50,  200, 10, False) # SafePeak
    26
    >>> bids, index = pm.merge_same_price(mar.bm.get_df())
    >>> mar.bm.get_df()
        quantity  price  user  buying  time  divisible
    0        250  200.0     0    True     0       True
    1        300  110.0     1    True     0       True
    2        120  100.0     2    True     0       True
    3         80   90.0     3    True     0       True
    4         40   85.0     4    True     0       True
    5         70   75.0     1    True     0       True
    6         60   65.0     0    True     0       True
    7         45   40.0     4    True     0       True
    8         30   38.0     3    True     0       True
    9         35   31.0     4    True     0       True
    10        25   24.0     0    True     0       True
    11        10   21.0     1    True     0       True
    12       120    0.0     5   False     0       True
    13        50    0.0     6   False     0       True
    14       200   15.0     7   False     0       True
    15       400   30.0     5   False     0       True
    16        60   32.5     8   False     0       True
    17        50   34.0     8   False     0       True
    18        60   36.0     8   False     0       True
    19       100   37.5     9   False     0       True
    20        70   39.0     9   False     0       True
    21        50   40.0     9   False     0       True
    22        70   60.0     5   False     0       True
    23        45   70.0     5   False     0       True
    24        50  100.0    10   False     0       True
    25        60  150.0    10   False     0       True
    26        50  200.0    10   False     0       True



    """

    id_gen = new_player_id(df.user.max() + 1)
    columns = df.columns.copy()

    df = df.copy().reset_index().rename(columns={'index': 'bid'})

    buy = df.loc[df['buying'], :]
    sell = df.loc[~df['buying'], :]

    dataframes = [buy, sell]

    agg_fun = {
        'bid': list,
        'user': list,
        'quantity': sum,
        'buying': lambda x: x.sample(1),
        'time': lambda x: x.sample(1),
        'divisible': lambda x: x.sample(1),
    }

    dataframe_new = []
    user_to_bid = {}
    for df_ in dataframes:
        rounded_prices = df_.price.apply(lambda x: np.round(x, prec))
        df_new = df_.groupby(rounded_prices).agg(agg_fun).reset_index()
        # print(df_new)
        df_new.user = df_new.user.apply(id_gen)
        #maping = df_new.set_index('user').bid.to_dict()
        # for k, v in maping.items():
        #    user_to_bid[k] = v

        dataframe_new.append(df_new)

    dataframe_new = pd.concat(dataframe_new).reset_index(drop=True)
    final_maping = dataframe_new.bid.to_dict()
    # print(final_maping)
    dataframe_new = dataframe_new[columns]
    # print('-------------')
    # print(dataframe_new)
    #index_to_user = dataframe_new.user.to_dict()

    #final_maping = {}
    # for k, v in index_to_user.items():
    #    final_maping[k] = user_to_bid[v]

    return dataframe_new, final_maping



from pymarket.bids import BidManager
from pymarket.mechanisms import Mechanism, MechanismReturn
from pymarket.transactions import TransactionManager
from pymarket.bids.demand_curves import *
from typing import List


def update_quantity(quantity: np.ndarray, gap: float) -> np.ndarray:
    """Implements the footnote in page 8 of [1],
    where the long side updates their
    trading quantities to match the short side.

    Parameters
    ----------
    quantity
        List of the quantities to be traded by each
        player.
    gap
        Difference between the short and long side

    Returns
    -------
    quantity
        Updated list of quantities to be traded
        by each player

    Notes
    ------
    [1] Huang, Pu, Alan Scheller–Wolf, and Katia Sycara. "Design of a multi–unit
    double auction e–market." Computational Intelligence 18.4 (2002): 596-617.

    Examples
    ---------
    All keep trading, with less quantity

    >>> l, g = np.array([1, 2, 3]), 0.6
    >>> update_quantity(l, g)
    array([0.8, 1.8, 2.8])

    The gap is to big for small trader:

    >>> l,g = np.array([1, 0.5, 2]), 1.8
    >>> update_quantity(l, g)
    array([0.35, 0.  , 1.35])

    """
    quantity = quantity * 1.0
    i_min = quantity[quantity > 0].argmin()
    v_min = quantity[quantity > 0].min()
    end = False
    N = len(quantity)
    while not end:
        if v_min < gap / N:
            quantity[i_min] = 0.0
            N -= 1
            gap -= v_min
            i_min = quantity[quantity > 0].argmin()
            v_min = quantity[quantity > 0].min()
        else:
            end = True
    quantity -= float(gap) / N
    max_ = quantity.max()
    quantity = np.clip(quantity, 0, max_)
    return quantity


def huang_auction(bids: pd.DataFrame) -> MechanismReturn:
    """Implements the auction described in [1]

    Parameters
    ----------
    bids:
        Collection of all the bids to take
        into account by the mechanism

    Returns
    -------
    trans : TransactionManager
        Collection of all the trasactions cleared
        by the mechanism

    extra : dict
        Extra information provided by the mecanism.
        Keys:
        * price_sell: price at which sellers traded
        * price_buy: price at which the buyers traded
        * quantity_traded: the total quantity traded


    Notes
    ------
    [1] Huang, Pu, Alan Scheller–Wolf, and Katia Sycara. "Design of a multi–unit
    double auction e–market." Computational Intelligence 18.4 (2002): 596-617.

    Examples
    --------
    No trade because price setters don't trade:

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(2, 1, 1)
    1
    >>> bm.add_bid(2, 2, 2, False)
    2
    >>> trans, extra = huang_auction(bm.get_df())
    >>> trans.get_df()
    Empty DataFrame
    Columns: [bid, quantity, price, source, active]
    Index: []
    >>> extra
    {'price_sell': 2.0, 'price_buy': 3.0, 'quantity_traded': 0}

    Adding small bids at the beginning, those can trade
    because they don't define de market price:

    >>> bm.add_bid(0.3, 1, 3, False)
    3
    >>> bm.add_bid(0.2, 3.3, 4)
    4
    >>> trans, extra = huang_auction(bm.get_df())
    >>> trans.get_df()
       bid  quantity  price  source  active
    0    3       0.2    2.0      -1   False
    1    4       0.2    3.0      -1   False
    >>> extra
    {'price_sell': 2.0, 'price_buy': 3.0, 'quantity_traded': 0.2}

    """

    # print(bids)

    trans = TransactionManager()

    buy, b_index = demand_curve_from_bids(bids)
    sell, s_index = supply_curve_from_bids(bids)

    q_, b_, s_, _ = intersect_stepwise(buy, sell)

    price_sell = sell[s_, 1]
    price_buy = buy[b_, 1]

    #quantity_buy = bids.iloc[b_index[:b_], 0].values
    #quantity_sell = bids.iloc[s_index[:s_], 0].values

    buying_bids = bids.loc[bids['buying']].sort_values(
        'price', ascending=False)
    selling_bids = bids.loc[~bids['buying']
                            ].sort_values('price', ascending=True)

    # Filter only the trading bids.
    buying_bids = buying_bids.iloc[: b_, :]
    selling_bids = selling_bids.iloc[: s_, :]

    # print(selling_bids, buying_bids)

    quantity_buy = buying_bids.quantity.values
    quantity_sell = selling_bids.quantity.values

    if b_ > 0 and s_ > 0:
        #long_sellers = sell[s_ - 1, 0] > buy[b_ - 1, 0]
        #gap = sell[s_ - 1, 0] - buy[b_ - 1, 0]
        gap = quantity_sell.sum() - quantity_buy.sum()
        if gap > 0:
            quantity_sell = update_quantity(quantity_sell, gap)
        else:
            quantity_buy = update_quantity(quantity_buy, - gap)

        for (i, x), q in zip(selling_bids.iterrows(), quantity_sell):
            # print(i)
            t = (i, q, price_sell, -1, False)
            trans.add_transaction(*t)

        for (i, x), q in zip(buying_bids.iterrows(), quantity_buy):
            # print(i)
            t = (i, q, price_buy, -1, False)
            trans.add_transaction(*t)

    #    for i in range(s_):
    #        id_ = s_index[i]
    #        trans.add_transaction(id_, quantity_sell[i],
    #                price_sell, -1, False)

    #    for i in range(b_):
    #        id_ = b_index[i]
    #        trans.add_transaction(id_, quantity_buy[i],
    #                price_buy, -1, False)

    extra = {'price_sell': price_sell, 'price_buy': price_buy,
             'quantity_traded': quantity_buy.sum()}
    # print(trans.get_df())
    return trans, extra


class HuangAuction(Mechanism):

    """Iinterface for the HuangAuction

    Parameters
    -----------
    bids: pd.DataFrame
        Collection of bids to use in the market
    merge: bool
        Wheather to merge players with the
        same price. Always `True`

    """

    def __init__(self, bids, *args, **kwargs):
        """
        """
        Mechanism.__init__(
            self,
            huang_auction,
            bids,
            *args,
            merge=True,
            **kwargs)



import numpy as np


def generate(
        cant_buyers: int,
        cant_sellers: int,
        offset_sellers: float=0,
        offset_buyers: float=0,
        r: np.random.RandomState = None,
        eps: float=1e-4):
    """
    Generates random bids. All the volumes and reservation
    prices are sampled independently from a uniform distribution.
    For sellers, the reservation price is shifted `offset_seller`
    while for the buyers is shifter `offset_buyers`.
    If there are two sellers or two buyers with the same price,
    the reservation price of one of them is resampled until
    in both side of the market, all players have different values.

    Parameters
    ----------
    cant_buyers: int
        Number of buyers to generate. Has to be positiv
    cant_sellers: int
        Number of sellers to generate. Has to be positive.
    offset_sellers
        Quantity to shift the reservation price of sellers
    offset_buyers
        Quantity to shift the reservation price of buyers
    r : optional
        RandomState used to generate the data
    eps : optional
        Minimum precision of the prices

    Returns
    -------
    bids
        List of tuples of all the bids generated

    Examples
    ---------
    >>> r = np.random.RandomState(420)
    >>> generate(2, 3, 1, 2, r, 0.1)
    [(0.5, 2.8, 0, True, 0, True), (0.7000000000000001, 2.5, 1, True, 0, True), (0.6000000000000001, 1.2, 2, False, 0, True), (0.1, 1.7000000000000002, 3, False, 0, True), (0.2, 1.3, 4, False, 0, True)]
    """

    if r is None:
        r = np.random.RandomState()

    offset = [offset_buyers, offset_sellers]
    quantities = [cant_buyers, cant_sellers]
    bids = []

    user = 0
    for i, (o_, q_) in enumerate(zip(offset, quantities)):
        range_ = np.arange(0, 1, eps)
        qs = r.choice(range_, q_, replace=False)
        vs = r.choice(range_ + o_, q_, replace=False)
        for j in range(q_):
            bid = (qs[j], vs[j], user, bool(1 - i), 0, True)
            bids.append(bid)
            user += 1
    return bids



import pandas as pd


class BidManager(object):
    """A class used to store and manipulate a collection
    of all the bids in the market.

    Attributes
    -----------
    col_names : :obj:`list` of :obj:`str`
        Column names for the different attributes in the dataframe
        to be created. Currently and in order: `quantity`, `price`,
        `user`, `buying`, `time`, `divisible`.
    n_bids : int
        Number of bids currently stored. Used as a unique identifier
        for each bid within a BidManager.
    bids : :obj:`list` of :obj:`tuple`
        A list where all the recieved bids are stored.
    """

    col_names = [
        'quantity',
        'price',
        'user',
        'buying',
        'time',
        'divisible',
    ]

    def __init__(self):
        self.n_bids = 0
        self.bids = []

    def add_bid(
        self,
        quantity: float,
        price: float,
        user: int,
        buying: bool=True,
        time: float = 0,
        divisible: bool =True
    ) -> int:
        """Appends a bid to the bid list

        Parameters
        ----------
        quantity
            Quantity of good desired. If `divisible=True` then any
            fraction of the good is an acceptable outcome of the
            market.
        price
            Uniform price offered in the market for each unit of the the good.
        user
            Identifier of the user submitting the bid.
        buying
            `True` if the bid is for buying the good and `False`otherwise.
            Default is `True`.
        time
            Instant at which the offer was made. This is relevant only if the
            market mechanism has perferences for earlier bids. Default is `0`
        divisible
            `True` is the user accepts a fraction of the asked quantity as
            a result and `False` otherwise.

        Returns
        -------
        int
            Unique identifier of the added bid.

        Examples
        --------
        >>> bm = pm.BidManager()
        >>> bm.add_bid(2, 1, 0)
        0
        """
        new_bid = (quantity, price, user, buying, time, divisible)
        self.bids.append(new_bid)
        self.n_bids += 1

        return self.n_bids - 1

    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe with the bids

        Parameters
        ----------

        Returns
        -------
        pd.DataFrame
            Dataframe with each row a different bid
            and each column each of the different attributes.

        Examples
        ---------
        >>> bm = pm.BidManager()
        >>> bm.add_bid(2, 1, 0)
        0
        >>> bm.add_bid(1, 3, 1, buying=False)
        1
        >>> bm.get_df()
           quantity  price  user  buying  time  divisible
        0         2      1     0    True     0       True
        1         1      3     1   False     0       True
        """

        df = pd.DataFrame(self.bids, columns=self.col_names)
        return df



# -*- coding: utf-8 -*-

"""Top-level package for pymarket."""

__author__ = """Diego Kiedanki"""
__email__ = 'gusok@protonmail.com'
__version__ = '0.5.3'

from pymarket.datasets.uniform_bidders import *



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from pymarket.bids.demand_curves import demand_curve_from_bids, supply_curve_from_bids


def plot_demand_curves(bids, ax=None, margin_X=1.2, margin_Y=1.2):
    """Plots the demand curves.
    If ax is none, creates a new figure

    Parameters
    ----------
    bids
          Collection of bids to be used

    ax : TODO, optional
         (Default value = None)
    margin_X :
         (Default value = 1.2)
    margin_Y :
         (Default value = 1.2)

    Returns
    -------


    """

    if ax is None:
        fig, ax = plt.subplots()

    extra_X = 3
    extra_Y = 1

    dc = demand_curve_from_bids(bids)[0]
    sp = supply_curve_from_bids(bids)[0]

    x_dc = dc[:, 0]
    x_dc = np.concatenate([[0], x_dc])
    x_sp = np.concatenate([[0], sp[:, 0]])

    y_sp = sp[:, 1]
    y_dc = dc[:, 1]
    max_x = max(x_dc[-2], x_sp[-2])
    extra_X = max_x * margin_X

    x_dc[-1] = extra_X
    y_dc = np.concatenate([y_dc, [0]])
    max_point = y_dc.max() * margin_Y

    x_sp[-1] = extra_X
    y_sp[-1] = max_point
    y_sp = np.concatenate([y_sp, [y_sp[-1]]])

    ax.step(x_dc, y_dc, where='post', c='r', label='Demand')
    ax.step(x_sp, y_sp, where='post', c='b', label='Supply')
    ax.set_xlabel('Quantity')
    ax.set_ylabel('Price')
    ax.legend()



from pymarket.mechanisms import *
from pymarket.bids import *
from pymarket.transactions import *

def test_huang_auction_with_dataset_0(bid_dataset_0):
    """Test the huang auction with dataset 0

    Parameters
    ----------
    bid_dataset_0 : TODO

    Returns
    -------
    TODO

    """

    df = bid_dataset_0
    trans, extra = huang_auction(df)
    
    df = trans.get_df().sort_values('bid').values.astype(float)

    true_trans = np.array([
        [1, 1, 3, -1, False],
        [4, 1, 2, -1, False]
        ]).astype(float)

    assert np.allclose(true_trans, df)

def test_huang_auction_with_dataset_3(bid_dataset_3):
    """Test the huang auction with dataset 3

    Parameters
    ----------
    bid_dataset_3 : TODO

    Returns
    -------
    TODO

    """

    df = bid_dataset_3
    trans, extra = huang_auction(df)
    
    df = trans.get_df().sort_values('bid').values.astype(float)

    true_trans = np.array([
        [0 , 0.8 , 3   , -1 , False] ,
        [1 , 0   , 3   , -1 , False] ,
        [2 , 1.7 , 3   , -1 , False] ,
        [4 , 1   , 2.5 , -1 , False] ,
        [5 , 1.5 , 2.5 , -1 , False]
        ]).astype(float)

    assert np.allclose(true_trans, df)



# -*- coding: utf-8 -*-

"""Top-level package for pymarket."""

__author__ = """Diego Kiedanki"""
__email__ = 'gusok@protonmail.com'
__version__ = '0.5.3'

from pymarket.mechanisms.mechanism import *
from pymarket.mechanisms.huang_auction import *
from pymarket.mechanisms.muda_auction import *
from pymarket.mechanisms.p2p_random import *



import pulp
import pandas as pd
from pymarket.bids import BidManager


def maximum_traded_volume(bids, *args, reservation_prices={}):
    """

    Parameters
    ----------
    bids : pd.DataFrame
       Collections of bids

    reservation_prices : dict of floats or None, (Default value = None)
        A maping from user ids to reservation prices. If no reservation
        price for a user is given, his bid will be assumed to be his
        true value.

    Returns
    -------
    status : str
        Status of the optimization problem. Desired output is 'Optimal'
    objective: float
        Maximum tradable volume that can be obtained
    variables: dict
        A set of values achieving the objective. Maps a pair of bids
        to the quantity traded by them.

    Examples
    ---------

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.add_bid(1.5, 1, 2, False)
    2
    >>> s, o, v = maximum_traded_volume(bm.get_df())
    >>> s
    'Optimal'
    >>> o
    1.5
    >>> v
    {(0, 2): 0.5, (1, 2): 1.0}

    """

    model = pulp.LpProblem("Max aggregated utility", pulp.LpMaximize)
    buyers = bids.loc[bids['buying']].index.values
    sellers = bids.loc[~bids['buying']].index.values

    index = [(i, j) for i in buyers for j in sellers
             if bids.iloc[i, 1] >= bids.iloc[j, 1]]

    qs = pulp.LpVariable.dicts('q', index, lowBound=0, cat='Continuous')

    model += pulp.lpSum([qs[x[0], x[1]] for x in index])

    for b in buyers:
        model += pulp.lpSum(qs[b, j]
                            for j in sellers if (b, j) in index) <= bids.iloc[b, 0]

    for s in sellers:
        model += pulp.lpSum(qs[i, s]
                            for i in buyers if (i, s) in index) <= bids.iloc[s, 0]

    model.solve()

    status = pulp.LpStatus[model.status]
    objective = pulp.value(model.objective)
    variables = {}
    for q in qs:
        v = qs[q].varValue
        variables[q] = v

    return status, objective, variables


def percentage_traded(bids, transactions, reservation_prices={}, **kwargs):
    """Calculates from the transaction dataframe
    the percentage of the total maximum possible
    traded quantity.

    Parameters
    ----------
    bids (pandas dataframe) :
        Table with all the submited bids
    transactions (pandas dataframe) :
        Table with all the transactions that ocurred in the market
    reservation_prices (dict, optional) :
        Reservation prices of the different participants. If None, the bids
        will be assumed to be the truthfull values.

    Returns
    -------
    ratio : float
        The ratio of the maximum social welfare achieved by the
        collection of transactions.

    Examples
    ---------
    Only bid 0 and 2 trade 1 unit. That represents
    the 66% of all that could have been traded.

    >>> tm = pm.TransactionManager()
    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.add_bid(1.5, 1, 2, False)
    2
    >>> tm.add_transaction(0, 1, 2, 2, False)
    0
    >>> tm.add_transaction(2, 1, 2, 0, False)
    1
    >>> percentage_traded(bm.get_df(), tm.get_df())
    0.6666666666666666

    """
    _, objective, _ = maximum_traded_volume(bids)

    total_traded = transactions.quantity.sum() / 2

    if objective > 0:
        return total_traded / objective
    else:
        return None



# -*- coding: utf-8 -*-

from pymarket.plot.demand_curves import *
from pymarket.plot.trades import *
from pymarket.plot.muda import *
from pymarket.plot.huang import *

"""Top-level package for pymarket."""

__author__ = """Diego Kiedanki"""
__email__ = 'gusok@protonmail.com'
__version__ = '0.5.3'



import numpy as np
from pymarket.statistics import *

def test_maxiumum_aggregated_utility_dataset_0(bid_dataset_0):

    df = bid_dataset_0
    status, objective, variables = maximum_aggregated_utility(df)
    assert status == 'Optimal'
    assert np.allclose(objective, 6)



import numpy as np
import pandas as pd
from typing import Tuple, Union
from pymarket.bids import BidManager


def demand_curve_from_bids(
        bids: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
    """
    Creates a demand curve from a set of buying bids.
    It is the inverse cumulative distribution of quantity
    as a function of price.

    Parameters
    ----------
    bids
        Collection of all the bids in the market. The algorithm
        filters only the buying bids.

    Returns
    ---------
    demand_curve: np.ndarray
       Stepwise constant demand curve represented as a collection
       of the N rightmost points of each interval (N-1 bids). It is stored
       as a (N, 2) matrix where the first column is the x-coordinate
       and the second column is the y-coordinate.
       An extra point is a))dded with x coordinate at infinity and
       price at 0 to represent the end of the curve.

    index : np.ndarray
        The order of the identifier of each bid in the demand
        curve.

    Examples
    ---------

    A minimal example, selling bid is ignored:

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 1, 0, buying=True)
    0
    >>> bm.add_bid(1, 1, 1, buying=False)
    1
    >>> dc, index = pm.demand_curve_from_bids(bm.get_df())
    >>> dc
    array([[ 1.,  1.],
           [inf,  0.]])
    >>> index
    array([0])

    A larger example with reordering of bids:

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 1, 0, buying=True)
    0
    >>> bm.add_bid(1, 1, 1, buying=False)
    1
    >>> bm.add_bid(3, 0.5, 2, buying=True)
    2
    >>> bm.add_bid(2.3, 0.1, 3, buying=True)
    3
    >>> dc, index = pm.demand_curve_from_bids(bm.get_df())
    >>> dc
    array([[1. , 1. ],
           [4. , 0.5],
           [6.3, 0.1],
           [inf, 0. ]])
    >>> index
    array([0, 2, 3])

    """
    buying = bids[bids.buying]
    buying = buying.sort_values('price', ascending=False)
    buying['acum'] = buying.quantity.cumsum()
    demand_curve = buying[['acum', 'price']].values
    demand_curve = np.vstack([demand_curve, [np.inf, 0]])
    index = buying.index.values
    return demand_curve, index


def supply_curve_from_bids(
        bids: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
    """
    Creates a supply curve from a set of selling bids.
    It is the cumulative distribution of quantity
    as a function of price.

    Parameters
    ----------
    bids
        Collection of all the bids in the market. The algorithm
        filters only the selling bids.

    Returns
    ---------
    supply_curve: np.ndarray
       Stepwise constant demand curve represented as a collection
       of the N rightmost points of each interval (N-1 bids). It is stored
       as a (N, 2) matrix where the first column is the x-coordinate
       and the second column is the y-coordinate.
       An extra point is added with x coordinate at infinity and
       price at infinity to represent the end of the curve.

    index : np.ndarray
        The order of the identifier of each bid in the supply
        curve.

    Examples
    ---------

    A minimal example, selling bid is ignored:

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0, False)
    0
    >>> bm.add_bid(2.1, 3, 3, True)
    1
    >>> sc, index = pm.supply_curve_from_bids(bm.get_df())
    >>> sc
    array([[ 1.,  3.],
           [inf, inf]])
    >>> index
    array([0])

    A larger example with reordering:

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0, False)
    0
    >>> bm.add_bid(2.1, 3, 3, True)
    1
    >>> bm.add_bid(0.2, 1, 3, False)
    2
    >>> bm.add_bid(1.7, 6, 4, False)
    3
    >>> sc, index = pm.supply_curve_from_bids(bm.get_df())
    >>> sc
    array([[0.2, 1. ],
           [1.2, 3. ],
           [2.9, 6. ],
           [inf, inf]])
    >>> index
    array([2, 0, 3])


    """
    selling = bids[bids.buying == False]
    selling = selling.sort_values('price')
    selling['acum'] = selling.quantity.cumsum()
    supply_curve = selling[['acum', 'price']].values
    supply_curve = np.vstack([supply_curve, [np.inf, np.inf]])
    index = selling.index.values
    return supply_curve, index


def get_value_stepwise(x: float, f: np.ndarray) -> Union[float, None]:
    """
    Returns the value of a stepwise constant
    function defined by the right extrems
    of its interval
    Functions are assumed to be defined
    in (0, inf).

    Parameters
    ----------
    x
        Value in which the function is to be
        evaluated
    f
        Stepwise function represented as a 2 column
        matrix. Each row is the rightmost extreme
        point of each constant interval. The first column
        contains the x coordinate and is sorted increasingly.
        f is assumed to be defined only in the interval
        :math: (0, \infty)
    Returns
    --------
    float or None
        The image of x under f: `f(x)`. If `x` is negative,
        then None is returned instead. If x is outside
        the range of the function (greater than `f[-1, 0]`),
        then the method returns None.

    Examples
    ---------
    >>> f = np.array([
    ...     [1, 1],
    ...     [3, 4]])
    >>> [pm.get_value_stepwise(x, f)
    ...     for x in [-1, 0, 0.5, 1, 2, 3, 4]]
    [None, 1, 1, 1, 4, 4, None]

    """
    if x < 0:
        return None

    for step in f:
        if x <= step[0]:
            return step[1]


def intersect_stepwise(
        f: np.ndarray,
        g: np.ndarray,
        k: float=0.5
    ) -> Tuple[
    Union[float, None],
    Union[int, None],
    Union[int, None],
        float]:
    """
    Finds the intersection of
    two stepwise constants functions
    where f is assumed to be bigger at 0
    than g.
    If no intersection is found, None is returned.

    Parameters
    ----------
    f
        Stepwise constant function represented as
        a 2 column matrix where each row is the rightmost
        point of the constat interval. The first column
        is sorted increasingly.
        Preconditions: f is non-increasing.

    g
        Stepwise constant function represented as
        a 2 column matrix where each row is the rightmost
        point of the constat interval. The first column
        is sorted increasingly.
        Preconditions: g is non-decreasing and
        `f[0, 0] > g[0, 0]`
    k
        If the intersection is empty or an interval,
        a convex combination of the y-values of f and g
        will be returned and k will be used to determine
        hte final value. `k=1` will be the value of g
        while `k=0` will be the value of f.

    Returns
    --------
    x_ast : float or None
        Axis coordinate of the intersection of both
        functions. If the intersection is empty,
        then it returns None.
    f_ast : int or None
        Index of the rightmost extreme
        of the interval of `f` involved in the
        intersection. If the intersection is
        empty, returns None
    g_ast : int or None
        Index of the rightmost extreme
        of the interval of `g` involved in the
        intersection. If the intersection is
        empty, returns None.
    v : float or None
        Ordinate of the intersection if it
        is uniquely identified, otherwise
        the k-convex combination of the
        y values of `f` and `g` in the last
        point when they were both defined.

    Examples
    ---------
    Simple intersection with diferent domains

    >>> f = np.array([[1, 3], [3, 1]])
    >>> g = np.array([[2,2]])
    >>> pm.intersect_stepwise(f, g)
    (1, 0, 0, 2)

    Empty intersection, returning the middle value

    >>> f = np.array([[1,3], [2, 2.5]])
    >>> g = np.array([[1,1], [2, 2]])
    >>> pm.intersect_stepwise(f, g)
    (None, None, None, 2.25)
    """
    x_max = np.min([f.max(axis=0)[0], g.max(axis=0)[0]])
    xs = sorted([x for x in set(g[:, 0]).union(set(f[:, 0])) if x <= x_max])
    fext = [get_value_stepwise(x, f) for x in xs]
    gext = [get_value_stepwise(x, g) for x in xs]
    x_ast = None
    for i in range(len(xs) - 1):
        if (fext[i] > gext[i]) and (fext[i + 1] < gext[i + 1]):
            x_ast = xs[i]

    f_ast = np.argmax(f[:, 0] >= x_ast) if x_ast is not None else None
    g_ast = np.argmax(g[:, 0] >= x_ast) if x_ast is not None else None

    g_val = g[g_ast, 1] if g_ast is not None else get_value_stepwise(xs[-1], g)
    f_val = f[f_ast, 1] if f_ast is not None else get_value_stepwise(xs[-1], f)

    intersect_domain_both = x_ast in f[:, 0] and x_ast in g[:, 0]
    if not (intersect_domain_both) and (x_ast is not None):
        v = g_val if x_ast in f[:, 0] else f_val
    else:
        v = g_val * k + (1 - k) * f_val

    return x_ast, f_ast, g_ast, v



import matplotlib.pyplot as plt
import numpy as np
from pymarket.plot.demand_curves import plot_demand_curves


def plot_huang_auction(bids, price_sell, price_buy, quantity_traded, ax=None):
    """Plots the results of the huang auction with some of the characteristics
    of such auction

    Parameters
    ----------
    bids (pandas dataframe):
        Table with all the bids submitted
    price_sell (list):
        The price at which all sellers sell
    price_buy (list):
        The price at which all players buy
    quantity traded (float):
        The total quantity traded

    Returns
    -------
    axe : matplotlib.axes._subplots.AxesSubplot
        The axe in which the figure was plotted.


    """
    bids = bids.get_df()
    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 8))
    plot_demand_curves(bids, ax=ax)
    ax.axhline(price_sell, linestyle='--', c='k', label='Sell price')
    ax.axhline(price_buy, linestyle='-.', c='k', label='Buy price')
    ax.axvline(quantity_traded, linestyle='--', c='k', label='Quantity traded')
    ax.fill_between(
        np.arange(
            0,
            quantity_traded,
            0.01),
        y1=price_buy,
        y2=price_sell,
        alpha=0.3,
        label='Market profit')
    ax.legend()

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = ['Click>=6.0', 'pandas>=0.24', 'numpy>=1.16', 'networkx', 'pulp', 'matplotlib']

setup_requirements = ['pytest-runner']

test_requirements = ['pytest']

setup(
    author="Diego Kiedanki",
    author_email='gusok@protonmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="A simple library for simulating markets in Python",
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='pymarket',
    name='pymarket',
    packages=['pymarket', 'pymarket.bids','pymarket.datasets', 'pymarket.mechanisms', 'pymarket.plot', 'pymarket.statistics', 'pymarket.transactions', 'pymarket.utils'],
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/gus0k/pymarket',
    version='0.7.5',
    zip_safe=False,
)



from pymarket.transactions import TransactionManager
from pymarket.mechanisms import *
import numpy as np


def test_example_1(bid_dataset_muda_example_1):
    """Test the example 1 presented in the MUDA
    paper

    Parameters
    ----------
    bid_dataset_muda_example_1 : Dataframe
        dataset with 3 biders from conftest

    Returns
    -------

    """
    df = bid_dataset_muda_example_1
    r = np.random.RandomState(1234)
    
    fees = np.zeros(df.user.unique().shape[0])
    trans = TransactionManager()
    price = 50
    trans, fees = solve_market_side_with_exogenous_price(df, price, fees)

    true_fees = np.array([0, 20, 10])
    assert np.allclose(true_fees, fees)

    data_trans = [
        [0, 1, 50, -1, False],
        [1, 1, 50, -1, False],
        [2, 1, 50, -1, False],
        [3, 1, 50, -1, False],
        [6, 1, 50, -1, False],
        [11, 1, 50, -1, False],
        [7, 1, 50, -1, False],
        [12, 1, 50, -1, False],
    ]
    # Exclude the active colum full of Nones
    data_trans = np.array(data_trans).astype(float)
    df2 = trans.get_df().values.astype(float)
    #print(data_trans)
    #print(df2)
    #assert 0 == 1
    assert np.allclose(data_trans, df2)


def test_muda_dataset_1(bid_dataset_1):
    """TODO: Testing muda with the splits

    Parameters
    ----------
    bid_dataset_1 : dataframe from conftest

    Returns
    -------

    """
    df = bid_dataset_1
    r = np.random.RandomState(1234)

    trans, extra = muda(df, r)

    fee = extra['fees']

    true_trans = np.array([
        [1, 1, 6.3, -1, False],
        [3, 1, 6.3, -1, False],
        [4, 1, 6.3, -1, False],
        [7, 1, 6.3, -1, False],
        [8, 2, 6.3, -1, False],
        [6, 1, 4.65, -1, False],
        [0, 1, 4.65, -1, False]]).astype(float)
    df = trans.get_df().values.astype(float)
    print(extra['left'], extra['right'])

    assert np.allclose(df, true_trans)
    
    left = [1, 3, 4, 7, 8, 9]
    right = [0, 2, 5, 6, 10]
    assert extra['left'] == left
    assert extra['right'] == right
    assert np.allclose(extra['price_left'], 4.65)
    assert np.allclose(extra['price_right'], 6.3)

    fees = np.array([1.85, 0, 0, 0, 0, 0, 0, 2.3, 4.6, 0, 0])
    assert np.allclose(fees, fee)




import matplotlib.pyplot as plt
from pymarket.plot.demand_curves import plot_demand_curves


def plot_both_side_muda(
        bids, left_players, right_players, left_price,
        right_price, FIGSIZE=(12, 6), **kwargs):
    """Plots the two sides in which MUDA divides the trades with the
    corresponding prices

    Parameters
    ----------
    bids (pandas dataframe):
        Table with all the bids submitted
    left (list):
        List of players in the left side
    right (list):
        List of players in the right side
    left_price (float):
        Price obtained from the left side to be used in the right side
    right_price (float):
        Price obtained from the right side to be used in the left side
    FIGSIZE (tuple):
        Tuple (width, height) of the figure to be created

    Returns
    -------
     axe : matplotlib.axes._subplots.AxesSubplot
        The axe in which the figure was plotted.
    """
    bids = bids.get_df()
    fig, ax = plt.subplots(1, 2, figsize=FIGSIZE)
    plot_demand_curves(bids[bids.index.isin(left_players)], ax=ax[0])
    ax[0].axhline(right_price, linestyle='--', c='k',
                  label='Trading price determined by right players')
    ax[0].legend()
    ax[0].set_title('Left players')

    plot_demand_curves(bids[bids.index.isin(right_players)], ax=ax[1])
    ax[1].axhline(
        left_price,
        linestyle='--',
        c='k',
        label='Trading price determined by left players')
    ax[1].legend()
    _ = ax[1].set_title('Right players')

    return ax



import numpy as np
import pandas as pd
from typing import Tuple, List
from pymarket.bids import BidManager, merge_same_price
from pymarket.transactions import TransactionManager, \
    split_transactions_merged_players
from pymarket.bids.demand_curves import *
from pymarket.mechanisms import Mechanism, MechanismReturn
from pymarket.utils.decorators import check_equal_price


@check_equal_price
def muda(bids: pd.DataFrame, r: np.random.RandomState=None) -> MechanismReturn:
    """Implements the Vickrey MUDA as described in [1].

    The mechanism does not support two players in the
    same side of the market with the same price.

    Parameters
    ----------
    bids
        Collection of bids to be used in the market
    r
        A numpy random state generator. If not given,
        a new one will be created and the output will
        be random.

    Returns
    -------
    trans: TransactionManager
        A collection of all the transactions performed.

    extra: dict
        Dictionary with extra information provided by
        the mechanism.
        Keys:
        * left: players in the left market
        * right: players in the right market
        * price_left: clearing price of the left market
        * price_right: clearing price of the right_market
        * fees: Fees that players have to pay to participate


    Notes
    ------

    [1] Segal-Halevi, Erel, Avinatan Hassidim, and Yonatan Aumann. "MUDA:
    a truthful multi-unit double-auction mechanism." Thirty-Second AAAI
    Conference on Artificial Intelligence. 2018.

    Examples
    ---------

    A case in which the market puts all the players
    in the same side and no one trades.

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 4, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.add_bid(1, 3, 2, False)
    2
    >>> bm.add_bid(1, 1, 3, False)
    3
    >>> r = np.random.RandomState(420)
    >>> trans, extra = muda(bm.get_df(), r)
    >>> extra
    {'left': [], 'right': [0, 1, 2, 3], 'price_left': inf, 'price_right': 2.5, 'fees': array([0., 0., 0., 0.])}
    >>> trans.get_df()
    Empty DataFrame
    Columns: [bid, quantity, price, source, active]
    Index: []

    A case in which there are 2 players in each side but the
    cleared prices makes it impossible to trade:

    >>> r = np.random.RandomState(69)
    >>> trans, extra = muda(bm.get_df(), r)
    >>> extra
    {'left': [1, 3], 'right': [0, 2], 'price_left': 1.5, 'price_right': 3.5, 'fees': array([0., 0., 0., 0.])}
    >>> trans.get_df()
    Empty DataFrame
    Columns: [bid, quantity, price, source, active]
    Index: []

    A case with trade:

    >>> bm.add_bid(1, 5, 4)
    4
    >>> r = np.random.RandomState(69)
    >>> trans, extra = muda(bm.get_df(), r)
    >>> trans.get_df()
       bid  quantity  price  source  active
    0    3         1    3.5      -1   False
    1    4         1    3.5      -1   False
    2    2         1    3.0      -1   False
    3    0         1    3.0      -1   False
    >>> extra
    {'left': [1, 3, 4], 'right': [0, 2], 'price_left': 3.0, 'price_right': 3.5, 'fees': array([0., 0., 0., 0., 0.])}

    """
    if r is None:
        r = np.random.RandomState()

    left = [i for i in bids.index if r.rand() > 0.5]
    right = [i for i in bids.index if i not in left]

    pl = find_competitive_price(bids.loc[left])
    pr = find_competitive_price(bids.loc[right])

    fees = np.zeros(bids.user.unique().shape[0])

    trans_left, fees = solve_market_side_with_exogenous_price(
        bids.loc[left], pr, fees)
    trans_right, fees = solve_market_side_with_exogenous_price(
        bids.loc[right], pl, fees)

    trans = trans_left.merge(trans_right)

    extra = {
        'left': left,
        'right': right,
        'price_left': pl,
        'price_right': pr,
        'fees': fees
    }
    return trans, extra


def solve_market_side_with_exogenous_price(
    bids: pd.DataFrame, price: float,
    fees: List[float]
) -> Tuple[TransactionManager, List[float]]:
    """
    Clears the market based on an external price.
    First it removes all biders that are not willing
    to trade at the given price, and then it fits
    the best allocation.
    Fees are calculated based on users that were
    willing but could not trade.

    Parameters
    ----------
    bids
        Collection of bids to clear the market with
    price
        Price at which all the trades will ocurr
    fees
        List of all the fees that players will have to pay.
        It gets updated.

    Returns
    -------
    trans:
        Collection of the transactions that clear the market
    fees:
        Fees to be paid by each player. Is a list where
        the fee of player with id `u` is located at `fees[u]`.

    Examples
    --------

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 0.5, 1)
    1
    >>> bm.add_bid(1, 1, 2, False)
    2
    >>> bm.add_bid(1, 2, 3, False)
    3
    >>> fees = [0, 0, 0, 0]
    >>> trans, fees = solve_market_side_with_exogenous_price(bm.get_df(),2.5, fees)
    >>> trans.get_df()
       bid  quantity  price  source  active
    0    0         1    2.5      -1   False
    1    2         1    2.5      -1   False
    >>> fees
    [0, 0, 0.5, 0]

    """
    trans = TransactionManager()
    demand = bids.loc[bids['buying']]
    if demand.shape[0] > 0:
        demand = demand.query('price >= @price')
        demand = demand.sort_values('price', ascending=False)
    else:
        return trans, fees
    supply = bids.loc[~bids['buying']]
    if supply.shape[0] > 0:
        supply = supply.query('price <= @price')
        supply = supply.sort_values('price')
    else:
        return trans, fees
    supply_quantity = supply.quantity.sum()
    demand_quantity = demand.quantity.sum()

    # Deal with the short side of the demand
    supply_long = supply_quantity > demand_quantity
    long_side = supply if supply_long else demand
    short_side = demand if supply_long else supply
    total_quantity = demand_quantity if supply_long else supply_quantity
    if total_quantity > 0:

        long_side, l_index = get_trading_bids(long_side, total_quantity)
        short_side, s_index = get_trading_bids(short_side, total_quantity)

        for i, x in short_side.iterrows():
            if i <= s_index:
                t = (x.bid, x.quantity, price, -1, False)
                trans.add_transaction(*t)

        for i, x in long_side.iterrows():
            if i <= l_index:
                t = (x.bid, x.quantity, price, -1, False)
                trans.add_transaction(*t)

        trading_users_long_side = (
            long_side[long_side.index <= l_index]
            .user
            .unique()
        )
        for u in trading_users_long_side:
            fee = compute_fee(long_side, l_index, u, total_quantity, price)
            fees[u] = fee

    return trans, fees


def get_trading_bids(
    bids: pd.DataFrame,
    quantity_traded: float
) -> Tuple[pd.DataFrame, int]:
    """
    Finds the index of the rightmost trading
    bid in a side of the market.
    If the bid has to be split, it does so, and
    returns the a new bid dataframe with two bids
    in stade of the original one.

    Parameters
    ----------
    bids
        Collection of bids in one side of the market
        Precondition: the dataframe is sorted by
        price. Reverse order for buying and selling
        side.

    quantity_traded
        Total quantity that the side of the market
        can trade.

    Returns
    -------
    bids_trading:
        Same as `bids`, but the index (which represent the
        bid identifier) is added as the first column.
        If a bid had to be splitted, that bid is replaced by two, with
        the quantity in both summing up to the original
        quantity. The index is reseted but both splitted
        bids retain the oringal bid number in the column.

    bid_index:
        Index of the `worst` bid that gets to trade.

    Examples
    ---------

    No splitting needed

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.get_df()
       quantity  price  user  buying  time  divisible
    0         1      3     0    True     0       True
    1         1      2     1    True     0       True
    >>> bids, index = get_trading_bids(bm.get_df(), 1)
    >>> bids
       bid  quantity  price  user  buying  time  divisible
    0    0         1      3     0    True     0       True
    1    1         1      2     1    True     0       True
    >>> index
    0

    Splitting needed:


    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 2, 1)
    1
    >>> bm.get_df()
       quantity  price  user  buying  time  divisible
    0         1      3     0    True     0       True
    1         1      2     1    True     0       True
    >>> bids, index = get_trading_bids(bm.get_df(), 0.3)
    >>> bids
       bid quantity price user buying time divisible
    0    0      0.3     3    0   True    0      True
    1    0      0.7     3    0   True    0      True
    2    1        1     2    1   True    0      True
    >>> index
    0

    """

    if bids.quantity.sum() > quantity_traded:
        bid_index = np.argmax(bids.quantity.cumsum().values >= quantity_traded)
    else:
        bid_index = bids.shape[0] - 1

    quantity_but_last = bids.iloc[: bid_index, :].quantity.sum()
    diff = quantity_traded - quantity_but_last
    if diff < bids.iloc[bid_index, :].quantity:
        new_row = pd.DataFrame(bids.iloc[bid_index, :]).T

        bids_trading = (pd.concat([
            bids.iloc[: bid_index, :],
            new_row,
            bids.iloc[bid_index:, :]
        ])
            .copy()
            .rename_axis('bid')
            .reset_index()
        )

        bids_trading.iloc[bid_index, 1] = diff
        bids_trading.iloc[bid_index + 1, 1] -= diff
    else:
        bids_trading = bids.copy().rename_axis('bid').reset_index()

    return bids_trading, bid_index


def compute_fee(
    df: pd.DataFrame,
    index: int,
    user: int,
    quantity: float,
    price: float
) -> float:
    """Computes the fee that a user has to pay by
    not letting others trade

    Parameters
    ----------
    df
        Dataframe for one side of the market
        resulting from reseting the index
        of a bid dataframe, getting the bid
        as the first column in addition to all the
        standard ones.
        Precondition: all bids should be willing
        to trade at the trading price.
    index
        Index of the last trading bid
    user
        User identifier for which the fee should be computed
    quantity
        Total quantity that the side of the
        market trades
    price
        Price at which the market clears.

    Returns
    -------
    fee
        Fee that user `ùser` will have to pay
        for not letting others trade as well.
    Examples
    --------

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 1, 1)
    0
    >>> bm.add_bid(1, 2, 3)
    1
    >>> compute_fee(bm.get_df(), 0, 1, 1, 2.5)
    0.5
    """
    trades_without_user = df[(df.user != user)]
    if trades_without_user.shape[0] == 0:
        fee = 0
    else:
        if trades_without_user.quantity.sum() > quantity:
            new_index = np.argmax(
                trades_without_user.quantity.cumsum().values >= quantity)
        else:
            new_index = trades_without_user.shape[0] - 1

        new_winning = trades_without_user.iloc[: new_index + 1, :].copy()
        diff = quantity - new_winning.iloc[:-1, :].quantity.sum()
        if diff < new_winning.iloc[new_index, :].quantity:
            new_winning.iloc[new_index, 1] = diff

        pure_new = new_winning[new_winning.index > index]
        fee = (pure_new.price - price) * pure_new.quantity
        fee = np.abs(fee.sum())
    return fee


def find_competitive_price(bids: pd.DataFrame) -> float:
    """
    Finds the unique trading price of the intersection
    of supply and demand.

    Parameters
    ----------
    bids
        Collection of bids to process the
        mechanism with.

    Returns
    -------
    price
        The price at which the market clears.

    Notes
    ------
    See also: intersect_stepwise.
    """

    buy, _ = demand_curve_from_bids(bids)
    sell, _ = supply_curve_from_bids(bids)

    q_, b_, s_, price = intersect_stepwise(buy, sell)

    return price


class MudaAuction(Mechanism):

    """Interface for MudaAuction.

    Parameters
    -----------
    bids
        Collection of bids to run the mechanism
        with.
    """

    def __init__(self, bids, *args, **kwargs):
        """TODO: to be defined1. """
        Mechanism.__init__(self, muda, bids, *args, **kwargs)



import pandas as pd

from pymarket.bids.processing import merge_same_price
from pymarket.transactions.processing import split_transactions_merged_players
from pymarket.transactions.transactions import TransactionManager
from typing import Tuple, Callable

MechanismReturn = Tuple[TransactionManager, dict]


class Mechanism():

    """Implements a standard interface for mechanisms

    Attributes
    -----------

    algo: Callable
        Algorithm to execute to solve the market.
    bids: pd.DataFrame
        Collection of bids to use, with processing.
    old_bids: pd.DataFrame
        Collection of bids previous to proecssing.
    maping: dict
        Map from the new bids to the old bids
    merge : bool
        Wheather to merge different players with
        the same price into one player. Useful for
        algorithms that require players to have different
        prices.

    Examples
    ---------

    Run p2p mechanism channging parameters with
    default parameters.

    >>> bm = pm.BidManager()
    >>> bm.add_bid(1, 3, 0)
    0
    >>> bm.add_bid(1, 0.5, 1)
    1
    >>> bm.add_bid(1, 1, 2, False)
    2
    >>> bm.add_bid(1, 2, 3, False)
    3
    >>> r = np.random.RandomState(420)
    >>> p2p = pm.mechanisms.p2p_random
    >>> mec = Mechanism(p2p, bm.get_df(), r=r)
    >>> trans, extra = mec.run()
    >>> extra
    {'trading_list': [[(0, 3), (1, 2)]]}
    >>> trans.get_df()
       bid  quantity  price  source  active
    0    0         1    2.5       3   False
    1    3         1    2.5       0   False
    2    1         0    0.0       2    True
    3    2         0    0.0       1    True

    >>> r = np.random.RandomState(420)
    >>> mec = Mechanism(p2p, bm.get_df(), r=r, p_coef=1)
    >>> trans, extra = mec.run()
    >>> extra
    {'trading_list': [[(0, 3), (1, 2)]]}
    >>> trans.get_df()
       bid  quantity  price  source  active
    0    0         1    3.0       3   False
    1    3         1    3.0       0   False
    2    1         0    0.0       2    True
    3    2         0    0.0       1    True
    """

    def __init__(self, algo, bids, *args, merge=False, **kwargs):
        """Creates a mechanisms with bids

        """
        self.algo = algo
        self.args = args
        self.kwargs = kwargs
        self.merge = merge
        self.bids = self._sanitize_bids(bids)

    def _sanitize_bids(self, bids: pd.DataFrame):
        """
        Method to be overwritten by different
        mechanisms according to their needs
        in preprocessing.

        Parameters
        ----------
        bids
            Collection of unprocess bids.

        Returns
        -------
        new_bids: pd.DataFrame
            The set of bids after processing.
        maping: dict
            Maping from new bids to old bids.


        Raises
        --------

        Examples
        ---------

        >>> bm = pm.BidManager()
        >>> bm.add_bid(1, 1, 0)
        0
        >>> bm.add_bid(0.5, 1, 1)
        1
        >>> mec = Mechanism(lambda: True, bm.get_df(), merge=True)
        >>> new_bids = mec._sanitize_bids(bm.get_df())
        >>> new_bids
           quantity  price  user  buying  time  divisible
        0       1.5      1   2.0     1.0   0.0        1.0
        >>> mec.maping
        {0: [0, 1]}
        """
        if self.merge:
            self.old_bids = bids
            new_bids, maping = merge_same_price(bids)
            # print(maping)
            self.maping = maping
        else:
            new_bids = bids

        return new_bids

    def _run(self):
        """Runs the mechanisms"""
        trans, extra = self.algo(self.bids, *self.args, **self.kwargs)
        return trans, extra

    def _cleanup(self, trans: pd.DataFrame):
        """Makes the necessary adjustements
        to return the transactions in a proper format.

        If players where merged, all transactios with
        the merged player are splited into the corresponding
        amount of transactions for the original players.

        Parameters
        ----------
        trans
            Collection of all the transactions executed
            in the market.


        Returns
        -------
        trans: pd.DataFrame
            The processed collection of transactions
            after the necesary adjustments.

        """

        if self.merge:
            trans = split_transactions_merged_players(
                trans, self.old_bids, self.maping)

        return trans

    def run(self):
        """Runs the mechanisms"""
        trans, extra = self._run()
        trans = self._cleanup(trans)
        return trans, extra
